// Copyright 2009, Gabriele Facciolo <gabriele.facciolo@upf.edu>
// Copyright 2009, Pablo Arias <pablo.arias@upf.edu>
// All rights reserved.

/*--------------------------- MegaWave2 module  -----------------------------
 name = mw_NLinterpVar
 authors = Gabriele Facciolo and Pablo Arias
 version = 0.2 (march 2014)
 function = "Exemplar-Based Interpolation of Sparsely Sampled Images", G. Facciolo, P. Arias, V. Caselles, and G. Sapiro, EMMCVPR, 2009.
 usage = 
   -s 1  : UNUSED
   -t 0  : tau (h in the paper)
   -i it : iterations
   -vid  : flag, if set stores a result at each iteration
   -N    : flag, if set averages the values of the pixels t>0
 	data     : data points 
   mask     : binary mask indicating the positions of the pixels
   initial  : initial condition for the interpolation
   out      : Interpolated image
   weight_out : optional weiths

*/
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include "fimage2.h"
#include "dimage.h"
#include "iio.h"

/*#define DEBUG_INPUT*/
#undef DEBUG_INPUT





extern void NLinject_discrete(long numpix, double *DX, double *DY, double *U, long newPix, double *OUTDX, double *OUTDY, double *OUTU, Dimage interp, Dimage dist, double param_tau, int NLmeans);

void mw_NLinterpVar(scale, tau,data,mask, interpolated, out,nlmeans,weight)
	Fimage data, mask, interpolated, out,weight;
     float scale;
	  float tau;
	  char nlmeans;
{
	long i,j,p;
	long nx,ny,dim;
	long numpix;
	double *DX,*DY,*U,*OUTDX,*OUTDY,*OUTU;
	Dimage interp,WEIG;

	nx = data->ncol;
	ny = data->nrow;
	numpix=0;
	for(i=0;i<nx;i++) {
		for(j=0;j<ny;j++) {
			if(mask->gray[i+j*nx] > 0) numpix++;
		}
	}

	DX = (double*) malloc(numpix*sizeof(double));
	DY = (double*) malloc(numpix*sizeof(double));
	U = (double*) malloc(numpix*sizeof(double));

	OUTDX = (double*) malloc(nx*ny*sizeof(double));
	OUTDY = (double*) malloc(nx*ny*sizeof(double));
	OUTU = (double*) malloc(nx*ny*sizeof(double));

	p=0;
	for(i=0;i<nx;i++) {
		for(j=0;j<ny;j++) {
			if(mask->gray[i+j*nx] > 0) {
				DX[p] = i;
				DY[p] = j;
				U[p] = data->gray[i+j*nx];
				p++;

				OUTU[i+j*nx] = data->gray[i+j*nx];
				OUTDX[i+j*nx] = i;
				OUTDY[i+j*nx] = j;
			} else{
				OUTU[i+j*nx] = -1000000.;
				OUTDX[i+j*nx] = i;
				OUTDY[i+j*nx] = j;
			}
			
		}
	}

	interp = new_dimage2(nx,ny);
	WEIG = new_dimage2(nx,ny);
	for(i=0;i<nx*ny;i++) { interp->gray[i] = interpolated->gray[i];} 
	
	NLinject_discrete(numpix, DX, DY, U, nx*ny, OUTDX, OUTDY, OUTU, interp, WEIG, tau, (int)nlmeans);

	for(i=0;i<nx*ny;i++) { out->gray[ (int)OUTDX[i] + (int)OUTDY[i]*nx ] = (float) OUTU[i]; }
	
	for(i=0;i<nx*ny;i++) { weight->gray[ (int)OUTDX[i] + (int)OUTDY[i]*nx ] = (float) WEIG->gray[i]; }

   del_dimage(interp);
   del_dimage(WEIG);

   free(DX);
   free(DY);
   free(U);

   free(OUTDX);
   free(OUTDY);
   free(OUTU);
}


/***************************/

// c: pointer to original argc
// v: pointer to original argv
// o: option name after hyphen
// d: default value (if NULL, the option takes no argument)
static char *pick_option(int *c, char ***v, char *o, char *d)
{
   int argc = *c;
   char **argv = *v;
   int id = d ? 1 : 0;
   for (int i = 0; i < argc - id; i++)
      if (argv[i][0] == '-' && 0 == strcmp(argv[i] + 1, o)) {
         char *r = argv[i + id] + 1 - id;
         *c -= id + 1;
         for (int j = i; j < argc - id; j++)
            (*v)[j] = (*v)[j + id + 1];
         return r;
      }
   return d;
}

/** IIO wrap. Read the file content into a new multi-channel Fimage
 * returns the allocated Fimage*/
Fimage fimageread(char *name)
{
   int nx, ny, nc;
   float *fdisp = iio_read_image_float_split(name, &nx, &ny, &nc);
   if (NULL == fdisp) {
      fprintf(stderr, "unreadable input files\n");
      exit(EXIT_FAILURE);
   }
   Fimage res = malloc_fimage3pla(nx, ny, nc);
   for (int i = 0; i < nx*ny*nc; i++)
      res->v[i] = fdisp[i];
   free(fdisp);
   return res;
}

/** IIO wrap. Write a file with the Fimage */
void fimagewrite(char *name, Fimage  i)
{
   if (i->planar == 1)
      iio_save_image_float_split(name, i->v, i->ncol, i->nrow, i->nch);
   else
      iio_save_image_float_vec(name, i->v, i->ncol, i->nrow, i->nch);
}



int main(int argc, char **argv)
{
   int iter =  atoi(pick_option(&argc, &argv, "i", "1"));
   float tau =  atof(pick_option(&argc, &argv, "t", "0"));
   float scale = atof(pick_option(&argc, &argv, "s", "1"));
   int nlmeans = pick_option(&argc, &argv, "N", NULL)?1:0;
   int vid = pick_option(&argc, &argv, "vid", NULL)?1:0;

   if (argc < 4) {
      fprintf(stderr, "too few parameters\n   %s [-i iter] [-t tau ] [-s scale] [-N] [-vid] data mask interp out [weight]\n", argv[0]);
      fprintf(stderr, " authors = Gabriele Facciolo and Pablo Arias\n");
      fprintf(stderr, " version = 0.2 (march 2014)\n");
      fprintf(stderr, " function = 'Exemplar-Based Interpolation of Sparsely Sampled Images, G. Facciolo, P. Arias, V. Caselles, and G. Sapiro, EMMCVPR, 2009.'\n");
      fprintf(stderr, " usage = \n");
      fprintf(stderr, "    data     : data points \n");
      fprintf(stderr, "    mask     : binary mask indicating the positions of the pixels\n");
      fprintf(stderr, "    initial  : initial condition for the interpolation\n");
      fprintf(stderr, "    out      : Interpolated image\n");
      fprintf(stderr, "    weight_out : (optional) final nonlocal weiths for each pixel\n");
      fprintf(stderr, "    -t 0  : tau (h in the paper)\n");
      fprintf(stderr, "    -i it : iterations\n");
      fprintf(stderr, "    -vid  : flag, if set stores a result at each iteration\n");
      fprintf(stderr, "    -N    : flag, if set averages the values of the pixels t>0\n");
      fprintf(stderr, "    -s 1  : UNUSED scale parameter\n");
 
      fprintf(stderr, "    ## PARAMETERS PROVUDED AS ENVIRONMENT VARIABLES\n");
      fprintf(stderr, "    ## Algorithm type is chosen by setting one of the following variables\n");
      fprintf(stderr, "       TYPE_AB=0 #algo. AB variational with Transmit and Receive (reccomended)\n");
      fprintf(stderr, "       TYPE_A=0  #algo. A. variational only with Receive\n");
      fprintf(stderr, "       TYPE_B=0  #algo. B. variational only with Transmit (reccomended for inpainting)\n");
      fprintf(stderr, "       TYPE_O=1  #algo. O. described in sect.4 'A departure from variational model' of the paper (default: as AB but faster)\n");
      fprintf(stderr, "   \n"); 
      fprintf(stderr, "    ## the patch support is detemined indirectly by the density of data samples\n");
      fprintf(stderr, "       NEIGHSIZE=49  #is the number of data samples used in a patch\n");
      fprintf(stderr, "       WINDOWSTD=12  #std of the spatial weights of the patch\n");
      fprintf(stderr, "   \n"); 
      fprintf(stderr, "    ## the non-local search is rerestricted to a locality of the current pixel\n");
      fprintf(stderr, "    ## the size of this locality is indirectly determined by the density of data samples\n");
      fprintf(stderr, "       WINSIZE=100   #the 100 nearest data samples are considered\n");
      fprintf(stderr, "       NOCENTER=1    #don't consider the center of the current patch as a neighbour (not very important)\n");
      fprintf(stderr, "   \n"); 
      fprintf(stderr, "    ## EXAMPLE CALL\n");
      fprintf(stderr, "       TYPE_AB=1 NOCENTER=1 WINSIZE=100 WINDOWSTD=12 NEIGHSIZE=49 ./mw_NLinterpVar -vid -i 150 -t 0 data.png mask.png init.png out.png weight_output.tif\n");

      return 1;
   }

   char *data_file = argv[1];
   char *mask_file = argv[2];
   char *interp_file = argv[3];
   char *out_file = argc > 4 ? argv[4] : NULL;
   char *out_weight_file = argc > 5 ? argv[5] : NULL;

   int ncol,nrow,nch;
   Fimage data = fimageread(data_file);
   ncol=data->ncol;nrow=data->nrow;nch=data->nch;
   Fimage mask = fimageread(mask_file);
   Fimage interpolated = fimageread(interp_file);

	/* Allocate thre results  */
   Fimage out = malloc_fimage3pla(ncol, nrow, nch);
   Fimage weight = malloc_fimage3pla(ncol, nrow, 1);

	/* Call the function  */
   for (int i=0;i<iter;i++) {
      float sumDIFF=0;
      mw_NLinterpVar(scale, tau,data,mask, interpolated, out,nlmeans,weight);
      for (int j=0;j<ncol*nrow*nch;j++) {
         if(mask->gray[j]>0) interpolated->gray[j]=data->gray[j];
         else                interpolated->gray[j]=out->gray[j];

         sumDIFF += fabs(interpolated->gray[j] - out->gray[j]);
      }
      printf("SUM absolute differences: %f\n",sumDIFF);
      // write evolution video
      if(vid){
         char tmpnam[2049];
         snprintf(tmpnam,  2048, "%s_%03d.png", out_file, i+1);
         fimagewrite(tmpnam, interpolated);
      }
   }

	/* write the results */
   fimagewrite(out_file, interpolated);
	if(out_weight_file) { 
      fimagewrite(out_weight_file, weight);
      free(weight);
	}
   free(out);
   free(data);
   free(mask);
   free(interpolated);
}


